package de.fhdw.wtf.common.ast;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;

import de.fhdw.wtf.common.ast.type.ClassType;
import de.fhdw.wtf.common.ast.type.ProductType;
import de.fhdw.wtf.common.ast.visitor.ConstructorOrOperationExceptionVisitor;
import de.fhdw.wtf.common.constants.parser.AstDescriptionConstants;
import de.fhdw.wtf.common.exception.walker.TaskException;
import de.fhdw.wtf.common.token.Token;

/**
 * This class represents a constructor of a class. It is held by a <ClassType>, can contain multiple
 * {@link ConstructorReference} of super class constructors and has parameters.
 */
public final class Constructor extends ConstructorOrOperation {
	
	/**
	 * The constructors called by this constructor.
	 * 
	 */
	private final Collection<ConstructorReference> superConstructors;
	
	/**
	 * Private-Constructor of {@link Constructor}.
	 * 
	 * 
	 * name of the constructor(generally the class-name of the constructor).
	 * 
	 * @param parameters
	 *            The parameters of the constructor.
	 * @param containingType
	 *            The classType of the class, that contains the constructor or operation.
	 * @param superConstructors
	 *            The constructors called by this constructor.
	 * @param firstToken
	 *            firstToken
	 * @param lastToken
	 *            lastToken
	 */
	private Constructor(final ProductType parameters,
			final ClassType containingType,
			final Collection<ConstructorReference> superConstructors,
			final Token firstToken,
			final Token lastToken) {
		super("", parameters, containingType, containingType, firstToken, lastToken);
		this.superConstructors = superConstructors;
	}
	
	/**
	 * Creates a {@link Constructor}-Object.
	 * 
	 * @param parameters
	 *            The parameters of the constructor.
	 * @param containingType
	 *            The classType of the class, that contains the constructor or operation.
	 * 
	 * @param superConstructors
	 *            The constructors called by this constructor
	 * @param firstToken
	 *            firstToken
	 * @param lastToken
	 *            lastToken
	 * @return The Constructor-Object
	 */
	public static Constructor create(final ProductType parameters,
			final ClassType containingType,
			final Collection<ConstructorReference> superConstructors,
			final Token firstToken,
			final Token lastToken) {
		return new Constructor(parameters, containingType, superConstructors, firstToken, lastToken);
		
	}
	
	/**
	 * Creates a {@link Constructor}-Object.
	 * 
	 * @param parameters
	 *            The parameters of the constructor.
	 * @param containingType
	 *            The classType of the class, that contains the constructor or operation.
	 * @param superConstructors
	 *            The constructors called by this constructor.
	 * @param firstToken
	 *            firstToken
	 * @return The Constructor-Object
	 */
	public static Constructor create(final ProductType parameters,
			final ClassType containingType,
			final Collection<ConstructorReference> superConstructors,
			final Token firstToken) {
		return new Constructor(parameters, containingType, superConstructors, firstToken, null);
		
	}
	
	@Override
	public String toString() {
		final StringBuilder result = new StringBuilder();
		result.append(this.getParameters());
		if (!this.superConstructors.isEmpty()) {
			result.append(AstDescriptionConstants.EQUAL_TOKEN);
			final Iterator<ConstructorReference> iterator = this.superConstructors.iterator();
			while (iterator.hasNext()) {
				result.append(iterator.next().toString());
				if (iterator.hasNext()) {
					result.append(AstDescriptionConstants.PLUS_TOKEN);
				}
			}
		}
		result.append(AstDescriptionConstants.SEMICOLON_TOKEN);
		return result.toString();
	}
	
	/**
	 * Represent the {@link Constructor}-Object as a String without the constructors called by this constructor.
	 * 
	 * @return {@link Constructor}-Object as a String
	 */
	public String toStringWithoutSuperconstructors() {
		final StringBuilder result = new StringBuilder();
		result.append(this.getParameters());
		return result.toString();
	}
	
	@Override
	public boolean equals(final Object o) {
		if (o instanceof Constructor) {
			final Constructor other = (Constructor) o;
			return this.superConstructors.containsAll(other.getSuperConstructors())
					&& other.getSuperConstructors().containsAll(this.superConstructors)
					&& this.getParameters().equals(other.getParameters());
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		int result = this.getParameters().hashCode();
		for (final ConstructorReference superConstructor : this.superConstructors) {
			result ^= superConstructor.hashCode();
		}
		return result;
	}
	
	/**
	 * The superConstructor of the constructor.
	 * 
	 * @return superConstructors
	 */
	public Collection<ConstructorReference> getSuperConstructors() {
		return Collections.unmodifiableCollection(this.superConstructors);
	}
	
	@Override
	public void accept(final ConstructorOrOperationExceptionVisitor<TaskException> constructorAndOperationVisitor)
			throws TaskException {
		constructorAndOperationVisitor.handleConstructor(this);
		
	}
}
