package de.fhdw.wtf.common.ast;

import java.util.Collection;
import java.util.Iterator;

import de.fhdw.wtf.common.ast.type.ProductElementType;
import de.fhdw.wtf.common.ast.type.ProductType;
import de.fhdw.wtf.common.ast.type.Type;
import de.fhdw.wtf.common.ast.visitor.ConstructorReferenceStateVisitor;
import de.fhdw.wtf.common.ast.visitor.ConstructorReferenceStateVisitorException;
import de.fhdw.wtf.common.ast.visitor.ConstructorReferenceStateVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.ConstructorReferenceStateVisitorReturnException;
import de.fhdw.wtf.common.constants.parser.AstDescriptionConstants;

/**
 * This class represents a state of a {@link ConstructorReference}, which already has been successfully referenced by
 * the Referencer.
 */
public final class ConstructorByReferenceState implements ConstructorReferenceState {
	
	/**
	 * The target-Constructor.
	 */
	private final Constructor target;
	
	/**
	 * The type of the target.
	 */
	private final Type targetType;
	
	/**
	 * Constructor of {@link ConstructorByReferenceState}.
	 * 
	 * @param constructor
	 *            the target-Constructor
	 * @param targetType
	 *            the type of the target.
	 */
	private ConstructorByReferenceState(final Constructor constructor, final Type targetType) {
		this.target = constructor;
		this.targetType = targetType;
	}
	
	/**
	 * Factory Method for {@link ConstructorByReferenceState}.
	 * 
	 * @param constructor
	 *            constructor
	 * @param targetType
	 *            targetType
	 * @return {@link ConstructorByReferenceState}
	 */
	public static ConstructorByReferenceState create(final Constructor constructor, final Type targetType) {
		return new ConstructorByReferenceState(constructor, targetType);
	}
	
	/**
	 * Returns the constructor that is the target of this.
	 * 
	 * @return the constructor that is the target of this.
	 */
	public Constructor getTarget() {
		return this.target;
	}
	
	/**
	 * Returns the type that the referenced constructor (target) belongs to.
	 * 
	 * @return the type that the referenced constructor (target) belongs to.
	 */
	public Type getType() {
		return this.targetType;
	}
	
	@Override
	public boolean equals(final Object o) {
		if (o instanceof ConstructorByReferenceState) {
			final ConstructorByReferenceState other = (ConstructorByReferenceState) o;
			final Constructor myTarget = this.getTarget();
			final Constructor otherTarget = other.getTarget();
			final ProductType myProductType = myTarget.getParameters();
			final ProductType otherProductType = otherTarget.getParameters();
			final Collection<ConstructorReference> myReferences = myTarget.getSuperConstructors();
			final Collection<ConstructorReference> otherReferences = otherTarget.getSuperConstructors();
			final Type myType = this.getType();
			final Type otherType = other.getType();
			return myType.getTypeString().equals(otherType.getTypeString()) && myProductType.equals(otherProductType)
					&& myReferences.containsAll(otherReferences) && otherReferences.containsAll(myReferences);
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		return this.getTarget().hashCode() ^ this.getType().getTypeString().hashCode();
	}
	
	@Override
	public String toString() {
		final StringBuilder result = new StringBuilder();
		result.append(AstDescriptionConstants.BRACKET_OPEN_TOKEN);
		final Iterator<ProductElementType> iterator = this.getTarget().getParameters().getElements().iterator();
		while (iterator.hasNext()) {
			final ProductElementType current = iterator.next();
			result.append(current.getType().getTypeString());
			if (iterator.hasNext()) {
				result.append(AstDescriptionConstants.COMMA_TOKEN);
			}
		}
		result.append(AstDescriptionConstants.BRACKET_CLOSE_TOKEN);
		return result.toString();
	}
	
	@Override
	public void accept(final ConstructorReferenceStateVisitor v) {
		v.handle(this);
	}
	
	@Override
	public <X> X accept(final ConstructorReferenceStateVisitorReturn<X> v) {
		return v.handle(this);
	}
	
	@Override
	public <Y extends Exception> void accept(final ConstructorReferenceStateVisitorException<Y> v) throws Y {
		v.handle(this);
	}
	
	@Override
	public <X, Y extends Exception> X accept(final ConstructorReferenceStateVisitorReturnException<X, Y> v) throws Y {
		return v.handle(this);
	}
}
