package de.fhdw.wtf.common.ast;

import de.fhdw.wtf.common.ast.visitor.NameReturnVisitor;
import de.fhdw.wtf.common.ast.visitor.NameVisitor;
import de.fhdw.wtf.common.ast.visitor.NameVisitorReturnBoolean;
import de.fhdw.wtf.common.ast.visitor.NameVisitorReturnName;
import de.fhdw.wtf.common.ast.visitor.NameVisitorReturnNameException;
import de.fhdw.wtf.common.exception.referencer.InvalidTypeReferenceException;
import de.fhdw.wtf.common.token.IdentifierToken;
import de.fhdw.wtf.common.token.Token;

/**
 * This class represents either a qualified name({@link QualifiedName}) or an unqualified name({@link UnqualifiedName}).
 * 
 */
public interface Name extends SyntaxObjectInterface {
	
	/**
	 * Add the identifierToken to the given hierarchy. Uses dummy-token as lastToken
	 * 
	 * @param identifierToken
	 *            new token to add.
	 * @return Name with added identifierToken
	 */
	Name addName(IdentifierToken identifierToken);
	
	/**
	 * Add the identifierToken to the given hierarchy.
	 * 
	 * @param lastToken
	 *            token to use as lastToken for Name
	 * @param identifierToken
	 *            new token to add.
	 * @return Name with added identifierToken
	 */
	Name addName(IdentifierToken identifierToken, Token lastToken);
	
	/**
	 * Visit-method for the {@link NameVisitorReturnName}-Visitor.
	 * 
	 * @param v
	 *            : NameVisitorForConcat
	 * @return : Name.
	 */
	Name visit(NameVisitorReturnName v);
	
	/**
	 * Visit-method for the {@link NameVisitor}-Visitor.
	 * 
	 * @param v
	 *            : NameVisitor.
	 * @throws InvalidTypeReferenceException
	 *             possible Exception
	 */
	void visit(NameVisitor v) throws InvalidTypeReferenceException;
	
	/**
	 * Visit-method for the {@link NameVisitorReturnBoolean}-Visitor.
	 * 
	 * @param v
	 *            : NameVisitor.
	 * @return Boolean
	 */
	boolean visit(NameVisitorReturnBoolean v);
	
	/**
	 * Visit-method for the {@link NameVisitorReturnNameException}-Visitor.
	 * 
	 * @param v
	 *            : NameVisitorException.
	 * @return Name
	 * @throws InvalidTypeReferenceException
	 *             possible Exception
	 */
	Name visit(NameVisitorReturnNameException<InvalidTypeReferenceException> v) throws InvalidTypeReferenceException;
	
	/**
	 * Visit-method for the {@link NameReturnVisitor}-Visitor.
	 * 
	 * @param v
	 *            Visitor
	 * @param <X>
	 *            ReturnType
	 * @return X
	 */
	<X> X visit(NameReturnVisitor<X> v);
	
	/**
	 * Concats <code>this</code> with the parameter <code>name</code>. Returns a new Name.
	 * 
	 * @param name
	 *            : Name.
	 * @return : Name.
	 */
	Name concat(Name name);
	
	/**
	 * This method returns true, if the last added name to <code>this</code> equals the first added part of
	 * <code>name</code>. False otherwise.
	 * 
	 * @param name
	 *            : Name.
	 * @return : boolean.
	 */
	boolean equalsPartial(Name name);
	
	/**
	 * Returns the last part of <code>this</code>.<br/>
	 * If <code>this</code> is an {@link UnqualifiedName} it will return itself.<br/>
	 * If <code>this</code> is a {@link QualifiedName} it will return the last UnqualifiedName of the QualifiedName.
	 * 
	 * @return : UnqualifiedName.
	 */
	UnqualifiedName getLastAddedName();
}
