package de.fhdw.wtf.common.ast.type;

import de.fhdw.wtf.common.ast.DatabaseIDState;
import de.fhdw.wtf.common.ast.Name;
import de.fhdw.wtf.common.ast.visitor.TypeProxyStateVisitor;
import de.fhdw.wtf.common.ast.visitor.TypeProxyStateVisitorException;
import de.fhdw.wtf.common.ast.visitor.TypeProxyStateVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.TypeProxyStateVisitorReturnException;

/**
 * This class represents a state of a {@link TypeProxy}, which already has been successfully referenced by the
 * Referencer.
 */
public final class ByReferenceState implements TypeProxyState {
	
	/**
	 * generated.
	 */
	private static final long serialVersionUID = -845179405519173802L;
	
	/**
	 * The target-Type.
	 */
	private final Type target;
	
	/**
	 * The target-Name. If the target is an {@link AtomicType} then the targetName should be the same as
	 * atomicType.getName().
	 */
	private final Name targetName;
	
	/**
	 * Constructor of {@link ByReferenceState}.
	 * 
	 * @param type
	 *            type
	 * @param targetName
	 *            targetName. If the target is an {@link AtomicType} then the targetName should be the same as
	 *            atomicType.getName().
	 */
	private ByReferenceState(final Type type, final Name targetName) {
		this.target = type;
		this.targetName = targetName;
	}
	
	/**
	 * Factory Method for {@link ByReferenceState}.
	 * 
	 * @param type
	 *            type
	 * @param targetName
	 *            targetName
	 * @return {@link ByReferenceState}
	 */
	public static ByReferenceState create(final Type type, final Name targetName) {
		return new ByReferenceState(type, targetName);
	}
	
	@Override
	public boolean equals(final Object o) {
		if (o instanceof Type) {
			if (!(o instanceof TypeProxy)) {
				final Type oAsType = (Type) o;
				return this.target.equals(oAsType);
			}
			return false;
		}
		if (o instanceof ByReferenceState) {
			final ByReferenceState other = (ByReferenceState) o;
			return other.getTargetName().equals(ByReferenceState.this.getTargetName());
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		return this.getTargetName().hashCode();
	}
	
	@Override
	public String toString() {
		// return this.getTarget().toString();
		// return ByReferenceState.class.getSimpleName();
		return this.getTargetName().toString();
	}
	
	@Override
	public String getTypeString() {
		return this.getTarget().getTypeString();
	}
	
	// //// PROJECTIONS //////
	
	/**
	 * Get the target-Type.
	 * 
	 * @return target-Type
	 */
	public Type getTarget() {
		return this.target;
	}
	
	/**
	 * Get the target-Name. If the target is an {@link AtomicType} then the targetName should be the same as
	 * atomicType.getName().
	 * 
	 * @return target-Name
	 */
	public Name getTargetName() {
		return this.targetName;
	}
	
	@Override
	public void accept(final TypeProxyStateVisitor v) {
		v.handle(this);
	}
	
	@Override
	public <X> X accept(final TypeProxyStateVisitorReturn<X> v) {
		return v.handle(this);
	}
	
	@Override
	public <Y extends Exception> void accept(final TypeProxyStateVisitorException<Y> v) throws Y {
		v.handle(this);
	}
	
	@Override
	public <X, Y extends Exception> X accept(final TypeProxyStateVisitorReturnException<X, Y> v) throws Y {
		return v.handle(this);
	}
	
	@Override
	public int compareTo(final TypeProxyState o) {
		return o.accept(new TypeProxyStateVisitorReturn<Integer>() {
			
			@Override
			public Integer handle(final ByNameState byName) {
				return 1;
			}
			
			@Override
			public Integer handle(final InvalidState invalidState) {
				return 1;
			}
			
			@Override
			public Integer handle(final ByReferenceState byType) {
				return byType.getTarget().compareTo(ByReferenceState.this.getTarget());
			}
		});
	}
	
	@Override
	public DatabaseIDState getTypeId() {
		return this.target.getTypeId();
	}
	
	@Override
	public void setTypeId(final DatabaseIDState typeId) {
		this.target.setTypeId(typeId);
	}
	
}
