package de.fhdw.wtf.common.task;

import java.util.ArrayList;
import java.util.List;

import de.fhdw.wtf.common.exception.walker.CyclicDependencyException;

/**
 * List of Dependencies for an "owner"-Task.
 * 
 */
public final class DependencyList {
	
	/**
	 * Dependencies.
	 */
	private final List<DependencyTask> dependencies;
	
	/**
	 * Owner.
	 */
	private final Task owner;
	
	/**
	 * Constructor of {@link DependencyList}.
	 * 
	 * @param owner
	 *            "owner"-Task.
	 */
	private DependencyList(final Task owner) {
		this.dependencies = new ArrayList<>();
		this.owner = owner;
	}
	
	/**
	 * Creates a {@link DependencyList}-Object.
	 * 
	 * @param owner
	 *            "owner"-Task.
	 * @return The {@link DependencyList}-Object.
	 */
	public static DependencyList create(final Task owner) {
		return new DependencyList(owner);
	}
	
	/**
	 * Adds a dependency.
	 * 
	 * @param task
	 *            dependency
	 * @throws CyclicDependencyException
	 *             possible Exception, if their are a cycle detected.
	 */
	public void addDependency(final DependencyTask task) throws CyclicDependencyException {
		if (task.getDependencies().containsDependency(this.owner)) {
			throw CyclicDependencyException.create(task.toString());
		}
		this.dependencies.add(task);
	}
	
	/**
	 * Returns true if the dependency list contains the given AbstractTask <code>task</code>.
	 * 
	 * @param task
	 *            : AbstractTask.
	 * @return boolean.
	 */
	public boolean containsDependency(final Task task) {
		boolean result = false;
		for (final DependencyTask current : this.dependencies) {
			if (task.equals(current)) {
				result = true;
			} else {
				result = result || current.getDependencies().containsDependency(task);
			}
		}
		return result;
	}
	
	/**
	 * Removes a dependency.
	 * 
	 * @param task
	 *            dependency
	 * @return True, if the dependency was remove.
	 */
	public boolean remove(final Task task) {
		return this.dependencies.remove(task);
	}
	
	/**
	 * returns true if the list of dependenies is empty, otherwise false.
	 * 
	 * @return Boolean
	 */
	public boolean isEmpty() {
		return this.dependencies.isEmpty();
	}
	
	/**
	 * Returns the list of dependencies.
	 * 
	 * @return dependencies
	 */
	public List<DependencyTask> getDependencies() {
		return this.dependencies;
	}
	
}
