package de.fhdw.wtf.dsl.scanner.modelScanner.state;

import java.util.regex.Pattern;

import de.fhdw.wtf.common.stream.TokenStream;
import de.fhdw.wtf.common.token.IdentifierToken;
import de.fhdw.wtf.common.token.Position;
import de.fhdw.wtf.common.token.keywords.AbstractToken;
import de.fhdw.wtf.common.token.keywords.ClassToken;
import de.fhdw.wtf.common.token.keywords.ExceptionToken;
import de.fhdw.wtf.common.token.keywords.FindableToken;
import de.fhdw.wtf.common.token.keywords.GroupToken;
import de.fhdw.wtf.common.token.keywords.MutableToken;
import de.fhdw.wtf.common.token.keywords.OperationToken;
import de.fhdw.wtf.common.token.keywords.PriorToken;
import de.fhdw.wtf.common.token.keywords.ServiceToken;
import de.fhdw.wtf.common.token.keywords.SymmetricToken;
import de.fhdw.wtf.common.token.keywords.TransientToken;
import de.fhdw.wtf.common.token.keywords.VisitableToken;

/**
 * ScannerState for recognition of multi-char Tokens such as KeywordToken or {@link IdentifierToken}.
 * 
 * @author tajoa
 */
public final class KeywordIdentifierState extends MultiCharState {
	
	/**
	 * Specifies that the given class is abstract.
	 */
	private static final String CLASS_MODI_ABSTRACT = "abstract";
	/**
	 * Specifies that the given identifier is typed as class.
	 */
	private static final String CLASS = "class";
	/**
	 * Specifies that the given identifier is typed as operation.
	 */
	private static final String OPERATION = "function";
	/**
	 * Specifies that the given identifier is typed as group.
	 */
	private static final String GROUP = "group";
	/**
	 * Specifies that the given identifier is typed as group.
	 */
	private static final String PRIOR = "prior";
	
	/**
	 * Specifies that the given class is an exception.
	 */
	private static final String EXCEPTION = "exception";
	
	/**
	 * Specifies that the given Attribute is findable.
	 */
	private static final String FINDABLE = "findable";
	
	/**
	 * Specifies that the given Attribute or Class is transient.
	 */
	private static final String TRANSIENT = "transient";
	
	/**
	 * Specifies that the given Class is a service.
	 */
	private static final String SERVICE = "service";
	
	/**
	 * Specifies that the given attribute is mutable.
	 */
	private static final String MUTABLE = "mutable";
	/**
	 * Specifies that for the given classes a visitor will be created.
	 */
	private static final String VISITABLE = "visitable";
	/**
	 * Specifies that for the given attribute a reverse getter will be created in the target type.
	 */
	private static final String SYMMETRIC = "symmetric";
	
	/**
	 * Specifies all allowed char.
	 */
	private static final String ALLOWED_CHARS = "\\w";
	
	/**
	 * <b>PRIVATE</b> Constructor.
	 * 
	 * @param currentInput
	 *            initializes currentInput
	 * @param firstPosition
	 *            initializes firstPostion
	 */
	private KeywordIdentifierState(final String currentInput, final Position firstPosition) {
		super(currentInput, firstPosition);
	}
	
	/**
	 * Factory for {@link KeywordIdentifierState}.
	 * 
	 * @param firstPosition
	 *            initializes firstPostion
	 * @return {@link KeywordIdentifierState}.
	 */
	public static KeywordIdentifierState create(final Position firstPosition) {
		return new KeywordIdentifierState("", firstPosition);
	}
	
	@Override
	public void finish(final TokenStream outputStream) {
		if (this.getCurrentInput().isEmpty()) {
			return;
		}
		if (this.getCurrentInput().equals(KeywordIdentifierState.CLASS)) {
			outputStream.add(ClassToken.create(this.getFirstPosition()));
		} else if (this.getCurrentInput().equals(KeywordIdentifierState.CLASS_MODI_ABSTRACT)) {
			outputStream.add(AbstractToken.create(this.getFirstPosition()));
		} else if (this.getCurrentInput().equals(KeywordIdentifierState.GROUP)) {
			outputStream.add(GroupToken.create(this.getFirstPosition()));
		} else if (this.getCurrentInput().equals(KeywordIdentifierState.PRIOR)) {
			outputStream.add(PriorToken.create(this.getFirstPosition()));
		} else if (this.getCurrentInput().equals(KeywordIdentifierState.EXCEPTION)) {
			outputStream.add(ExceptionToken.create(this.getFirstPosition()));
		} else if (this.getCurrentInput().equals(KeywordIdentifierState.FINDABLE)) {
			outputStream.add(FindableToken.create(this.getFirstPosition()));
		} else if (this.getCurrentInput().equals(KeywordIdentifierState.TRANSIENT)) {
			outputStream.add(TransientToken.create(this.getFirstPosition()));
		} else if (this.getCurrentInput().equals(KeywordIdentifierState.SERVICE)) {
			outputStream.add(ServiceToken.create(this.getFirstPosition()));
		} else if (this.getCurrentInput().equals(KeywordIdentifierState.OPERATION)) {
			outputStream.add(OperationToken.create(this.getFirstPosition()));
		} else if (this.getCurrentInput().equals(KeywordIdentifierState.MUTABLE)) {
			outputStream.add(MutableToken.create(this.getFirstPosition()));
		} else if (this.getCurrentInput().equals(KeywordIdentifierState.VISITABLE)) {
			outputStream.add(VisitableToken.create(this.getFirstPosition()));
		} else if (this.getCurrentInput().equals(KeywordIdentifierState.SYMMETRIC)) {
			outputStream.add(SymmetricToken.create(this.getFirstPosition()));
		} else {
			outputStream.add(IdentifierToken.create(this.getCurrentInput(), this.getFirstPosition()));
		}
	}
	
	/**
	 * Checks whether the input char is a valid identifier/keyword character or not.
	 * 
	 * @param input
	 *            character to be checked
	 * @return <code>true</code> if input is a allowed char. <br>
	 *         <code>false</code> if input is not allowed.
	 */
	public static boolean isAllowedCharacter(final char input) {
		return Pattern.matches(KeywordIdentifierState.ALLOWED_CHARS, Character.toString(input));
	}
	
	@Override
	public boolean accept(final ScannerStateVisitor visitor) {
		return visitor.handleKeywordIdentifierState();
	}
}
