package de.fhdw.wtf.parser;

import java.util.Collection;

import de.fhdw.wtf.common.ast.type.ProductType;
import de.fhdw.wtf.common.exception.parser.AbstractParserException;
import de.fhdw.wtf.common.exception.parser.NoBracketCloseException;
import de.fhdw.wtf.common.stream.TokenStream;
import de.fhdw.wtf.common.token.Token;

/**
 * Parser to parse the given TokenStream to {@link ProductType}s.
 */
final class ProductParser {
	
	/**
	 * TokenStream.
	 */
	private final TokenStream stream;
	
	/**
	 * Collection of exceptions.
	 */
	private final Collection<AbstractParserException> exceptions;
	
	/**
	 * Constructor of {@link ProductParser}.
	 * 
	 * @param stream
	 *            tokenStream
	 * @param exceptions
	 *            collection of exceptions
	 */
	private ProductParser(final TokenStream stream, final Collection<AbstractParserException> exceptions) {
		this.stream = stream;
		this.exceptions = exceptions;
	}
	
	/**
	 * FactoryMethod for {@link ProductParser}.
	 * 
	 * @param stream
	 *            tokenStream
	 * @param exceptions
	 *            collection of exceptions
	 * @return The {@link ProductParser}-Object.
	 */
	static ProductParser create(final TokenStream stream, final Collection<AbstractParserException> exceptions) {
		return new ProductParser(stream, exceptions);
	}
	
	/**
	 * Parses a series of tokens and creates a {@link ProductType} from it. If the series of tokens does not match the
	 * grammar an Exception will be thrown.
	 * 
	 * @return ProductType
	 * @throws AbstractParserException
	 *             AbstractParserException
	 */
	ProductType parse() throws AbstractParserException {
		final Token firstToken = this.stream.next();
		final ProductType prod = ProductType.create(firstToken);
		boolean nextTokenType = true;
		if (this.stream.peek().isBracketCloseToken()) {
			nextTokenType = false;
		}
		while (this.stream.hasNext() && this.stream.peek().isIdentifierToken() && nextTokenType) {
			final ParameterParser parameterParser = ParameterParser.create(this.stream, this.exceptions);
			prod.addElement(parameterParser.parse());
			if (this.stream.peek().isCommaToken()) {
				this.stream.removeFirst();
				nextTokenType = true;
			} else {
				nextTokenType = false;
			}
		}
		if (this.stream.hasNext() && this.stream.peek().isBracketCloseToken()) {
			this.stream.removeFirst();
		} else {
			throw NoBracketCloseException.create(this.stream.peek());
		}
		prod.setLastToken(this.stream.peek());
		return prod;
	}
}
