package de.fhdw.wtf.tooling;

import java.io.IOException;
import java.net.URI;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collection;
import java.util.concurrent.ExecutionException;

import de.fhdw.wtf.common.ast.Model;
import de.fhdw.wtf.common.exception.editor.CheckException;
import de.fhdw.wtf.common.exception.editor.EditorMarkable;
import de.fhdw.wtf.common.exception.editor.EditorMarker;
import de.fhdw.wtf.common.exception.editor.GeneralCheckException;
import de.fhdw.wtf.common.exception.editor.MarkableCheckException;
import de.fhdw.wtf.common.exception.editor.MultipleCheckExceptions;
import de.fhdw.wtf.common.exception.parser.AbstractParserException;
import de.fhdw.wtf.common.exception.parser.NoValidTokenStreamException;
import de.fhdw.wtf.common.exception.referencer.ReferencerException;
import de.fhdw.wtf.common.exception.referencer.ReferencingNotSuccessfulException;
import de.fhdw.wtf.common.exception.walker.CyclicDependencyException;
import de.fhdw.wtf.common.exception.walker.CyclicPartDefinitionException;
import de.fhdw.wtf.common.exception.walker.TaskException;
import de.fhdw.wtf.common.stream.FilteredTokenStream;
import de.fhdw.wtf.common.stream.SimpleScannerInput;
import de.fhdw.wtf.common.task.DependencyTask;
import de.fhdw.wtf.common.task.GroupDependencyTask;
import de.fhdw.wtf.common.task.TaskExecutorFixed;
import de.fhdw.wtf.common.task.result.ExceptionalTaskResult;
import de.fhdw.wtf.common.task.result.OKTaskResult;
import de.fhdw.wtf.common.task.result.TaskResult;
import de.fhdw.wtf.common.task.result.visitor.TaskResultVisitor;
import de.fhdw.wtf.common.token.Position;
import de.fhdw.wtf.dsl.scanner.modelScanner.ModelDslScanner;
import de.fhdw.wtf.parser.Parser;
import de.fhdw.wtf.walker.tasks.AbstractOperationsCheck;
import de.fhdw.wtf.walker.tasks.AnalyzeInheritanceTreesTask;
import de.fhdw.wtf.walker.tasks.BaseTypeInheritanceCheck;
import de.fhdw.wtf.walker.tasks.ConstructorReferencer;
import de.fhdw.wtf.walker.tasks.CyclicInheritanceCheck;
import de.fhdw.wtf.walker.tasks.DoubleAttributenameCheck;
import de.fhdw.wtf.walker.tasks.DoubleGroupcomponentCheck;
import de.fhdw.wtf.walker.tasks.InvalidAttributeModifierCheck;
import de.fhdw.wtf.walker.tasks.OverloadingCheck;
import de.fhdw.wtf.walker.tasks.ProductCheck;
import de.fhdw.wtf.walker.tasks.PrototypesTask;
import de.fhdw.wtf.walker.tasks.ReferencedConstructorsCheck;
import de.fhdw.wtf.walker.tasks.SubtypesFillTask;
import de.fhdw.wtf.walker.tasks.SupertypesCleanInheritanceTask;
import de.fhdw.wtf.walker.tasks.SupertypesFillTask;
import de.fhdw.wtf.walker.tasks.SymmetricCheck;
import de.fhdw.wtf.walker.tasks.TypeReferencer;
import de.fhdw.wtf.walker.walker.SimpleWalkerTask;

public abstract class SyntaxCheckAbstract {
	
	private static final Charset DEFAULT_CHARSET = StandardCharsets.UTF_8;
	
	/**
	 * Erzeugt ein Modell aus dem übergebenen Pfad zum WTF-Modell.
	 * 
	 * @param uri
	 * @return
	 * @throws MultipleCheckExceptions
	 * @throws CyclicDependencyException
	 * @throws InterruptedException
	 * @throws CyclicPartDefinitionException
	 * @throws ExecutionException
	 */
	public Model getModelFromFile(final URI uri) throws MultipleCheckExceptions, CyclicDependencyException,
			InterruptedException, CyclicPartDefinitionException, ExecutionException {
		try {
			final String fileContent = readFile(uri, DEFAULT_CHARSET);
			return this.getModelFromString(fileContent);
		} catch (final IOException e) {
			e.printStackTrace();
			return null;
		}
	}
	
	/**
	 * Erzeugt ein Modell aus dem übergebenen Inhalt einer Datei die ein WTF-Modell enthält.
	 * 
	 * @param wtfFileContent
	 * @return
	 * @throws MultipleCheckExceptions
	 * @throws CyclicDependencyException
	 * @throws InterruptedException
	 * @throws CyclicPartDefinitionException
	 * @throws ExecutionException
	 */
	public Model getModelFromString(final String wtfFileContent) throws MultipleCheckExceptions,
			CyclicDependencyException, InterruptedException, CyclicPartDefinitionException, ExecutionException {
		final FilteredTokenStream output = this.scanString(wtfFileContent);
		Model result;
		result = this.parseStream(output);
		this.checkModelGeneral(result);
		return result;
	}
	
	/**
	 * Internal Method. This Method check the model.
	 *
	 * @param model
	 */
	public void checkModelGeneral(final Model model) throws MultipleCheckExceptions, CyclicDependencyException,
			InterruptedException, ExecutionException, CyclicPartDefinitionException {
		final MultipleCheckExceptions exceptions = new MultipleCheckExceptions();
		final TaskExecutorFixed taskmanager = TaskExecutorFixed.create();
		
		final GroupDependencyTask firstCheckers = GroupDependencyTask.create(taskmanager);
		final DependencyTask doubleGroupcomponent = DoubleGroupcomponentCheck.create(model, taskmanager);
		final DependencyTask abstractOperationsCheck = AbstractOperationsCheck.create(model, taskmanager);
		final DependencyTask productCheck = ProductCheck.create(model, taskmanager);
		final DependencyTask symmetricCheck = SymmetricCheck.create(model, taskmanager);
		firstCheckers.addMembers(doubleGroupcomponent, abstractOperationsCheck, productCheck, symmetricCheck);
		
		final TypeReferencer typeReferencer = TypeReferencer.create(model, taskmanager);
		typeReferencer.addDependency(firstCheckers);
		
		final GroupDependencyTask inheritanceChecker = GroupDependencyTask.create(taskmanager);
		final DependencyTask cyclicInheritance = CyclicInheritanceCheck.create(model, taskmanager);
		final DependencyTask baseTypeInheritanceCheck = BaseTypeInheritanceCheck.create(model, taskmanager);
		inheritanceChecker.addMembers(cyclicInheritance, baseTypeInheritanceCheck);
		inheritanceChecker.addDependency(typeReferencer);
		
		final GroupDependencyTask checkersAfterReferencing = GroupDependencyTask.create(taskmanager);
		final SimpleWalkerTask invalidAttributeModifierCheck = InvalidAttributeModifierCheck.create(model, taskmanager);
		final DependencyTask doubleAttributenameCheck = DoubleAttributenameCheck.create(model, taskmanager);
		checkersAfterReferencing.addMembers(invalidAttributeModifierCheck, doubleAttributenameCheck);
		checkersAfterReferencing.addDependency(inheritanceChecker);
		
		final GroupDependencyTask completers = GroupDependencyTask.create(taskmanager);
		final DependencyTask prototypesTask = PrototypesTask.create(model, taskmanager);
		final DependencyTask subtypesFillTask = SubtypesFillTask.create(model, taskmanager);
		final SupertypesFillTask supertypesFillTask = SupertypesFillTask.create(model, taskmanager);
		final DependencyTask supertypesCleanInheritanceTask =
				SupertypesCleanInheritanceTask.create(model, taskmanager, supertypesFillTask);
		prototypesTask.addDependency(subtypesFillTask);
		supertypesFillTask.addDependency(prototypesTask);
		supertypesCleanInheritanceTask.addDependency(supertypesFillTask);
		completers.addMembers(subtypesFillTask, prototypesTask, supertypesFillTask, supertypesCleanInheritanceTask);
		completers.addDependency(checkersAfterReferencing);
		final SimpleWalkerTask overloadingCheck = OverloadingCheck.create(model, taskmanager);
		overloadingCheck.addDependency(completers);
		
		final ConstructorReferencer constructorReferencer = ConstructorReferencer.create(model, taskmanager);
		constructorReferencer.addDependency(overloadingCheck);
		
		final DependencyTask analyzeInheritanceTreesTask = AnalyzeInheritanceTreesTask.create(model, taskmanager);
		analyzeInheritanceTreesTask.addDependency(constructorReferencer);
		
		final DependencyTask referencedConstructorsCheck = ReferencedConstructorsCheck.create(model, taskmanager);
		referencedConstructorsCheck.addDependency(analyzeInheritanceTreesTask);
		
		taskmanager.startAllKnownTasks();
		final Collection<TaskResult> results = taskmanager.getResultsAndShutdown();
		
		final Collection<Exception> occuredExceptions = new ArrayList<>();
		for (final TaskResult current : results) {
			current.accept(new TaskResultVisitor() {
				
				@Override
				public void handleOkTaskResult(final OKTaskResult result) {
					// Everything right
				}
				
				@Override
				public void handleExceptionalTaskResult(final ExceptionalTaskResult result) {
					occuredExceptions.add(result.getError());
				}
			});
		}
		if (!occuredExceptions.isEmpty()) {
			for (final Exception occuredException : occuredExceptions) {
				if (occuredException instanceof ReferencingNotSuccessfulException) {
					for (final TaskException e : typeReferencer.getExceptions()) {
						if (e instanceof ReferencerException) {
							exceptions.add(new MarkableCheckException(((ReferencerException) e).getPosStart(),
									((ReferencerException) e).getPosEnd(), e, EditorMarker.REFERENCER_MARKER));
						}
					}
				} else if (occuredException instanceof EditorMarkable) {
					final EditorMarkable e = (EditorMarkable) occuredException;
					exceptions.add(new MarkableCheckException(e.getStartPosition(), e.getEndPos(), occuredException, e
							.getMarker()));
				} else {
					exceptions.add(new GeneralCheckException(occuredException));
				}
			}
			
			throw exceptions;
		}
	}
	
	private Model parseStream(final FilteredTokenStream output) throws MultipleCheckExceptions {
		Model result;
		final Parser parser = Parser.create(output);
		try {
			result = parser.parse();
			return result;
		} catch (final NoValidTokenStreamException e) {
			
			final Collection<CheckException> excList = new ArrayList<>();
			for (final AbstractParserException parserException : parser.getExceptions()) {
				final int positionInLine =
						parserException.getToken().getPosition().getPosition() + parserException.getToken().getLength();
				final int columnNumber =
						parserException.getToken().getPosition().getPosition() + parserException.getToken().getLength();
				final int lineNumber = parserException.getToken().getPosition().getLineNumber();
				final Position startPos = parserException.getToken().getPosition();
				final Position endPos =
						Position.create(
								parserException.getToken().getPosition().getFilePath(),
								lineNumber,
								columnNumber,
								positionInLine);
				final MarkableCheckException markableException =
						new MarkableCheckException(startPos, endPos, parserException, EditorMarker.PARSER_MARKER);
				excList.add(markableException);
			}
			throw new MultipleCheckExceptions(excList); // NOPMD - "e" is not a cause here
		}
	}
	
	private FilteredTokenStream scanString(final String fileContent) {
		final FilteredTokenStream output = FilteredTokenStream.create();
		final ModelDslScanner scanner = ModelDslScanner.create();
		final SimpleScannerInput input = new SimpleScannerInput(fileContent);
		scanner.scan(input, output);
		return output;
	}
	
	private static String readFile(final URI uri, final Charset encoding) throws IOException {
		final byte[] encoded = Files.readAllBytes(Paths.get(uri.getPath()));
		return new String(encoded, encoding);
	}
	
}
